﻿// <summary>ソースコード：muskingumモデルクラス</summary>
// <author>CommonMP</author>

using System;
using System.Collections.Generic;
using System.Collections;
using System.Text;
using System.Diagnostics;

using CommonMP.HYSSOP.CoreImpl.HSData;
using CommonMP.HYSSOP.CoreImpl.HSTools;

using CommonMP.HYMCO.Interface.Model;
using CommonMP.HYMCO.Interface.Data;
using CommonMP.HYMCO.CoreImpl.Data;
using CommonMP.HYMCO.CoreImpl.Tool;



namespace jp.or.jccajsce.MuskingumModel
{

    /// <summary><para>class outline:</para>
    /// <para>muskingumモデル</para>
    /// </summary>
    /// <remarks><para>history:</para>
    /// <para>[CommonMP][ver 1.0.0][2010/01][新規作成]</para>
    /// </remarks>
    public class MuskingumModel : McStateCalModelBase
    {
        //=========================
        // 演算実行前処理関連メソッド
        //=========================

        ///// <summary><para>method outline:</para>
        ///// <para>計算開始時に動作する処理</para>
        ///// </summary>
        ///// <example><para>usage:</para>
        ///// <para>bool bRtn = ReadyCalculation( )</para>
        ///// </example>
        ///// <param name="">無し</param>
        ///// <returns>true :正常 , false:異常</returns>
        ///// <exception cref="">無し</exception>
        ///// <remarks><para>remarks:</para>
        ///// <para>計算開始の　最初に１回だけコールされるメソッドです。不要ならば、本メソッドをオーバーライドする必要はありません。</para>
        ///// </remarks>
        //public override bool ReadyCalculation()

        /// <summary><para>method outline:</para>
        /// <para>入力側の接続情報チェック</para>
        /// </summary>
        /// <example><para>usage:</para>
        /// <para>bool bRtn = ReceiveConnectionCheck(ref csErrorInf)</para>
        /// </example>
        /// <param name="csErrorInf">エラー出力</param>
        /// <param name="lInputDataNum">入力情報数</param>
        /// <param name="csInputCellData">演算に必要な入力情報配列</param>
        /// <returns>=true:正常、=false:異常</returns>
        /// <exception cref="">無し</exception>
        /// <remarks><para>remarks:</para>
        /// <para>受信するデータが自モデルが期待している情報か否かをチェックする</para>
        /// </remarks>
        protected override bool ReceiveConnectionCheck(ref McStructErrorInfo csErrorInf, long lInputDataNum, McReceiveCellDataIF[] csInputCellData)
        {
            bool bRtn = true;

            // 入力端子数のチェック
            if (lInputDataNum == 0)
            {
                csErrorInf.AddCheckErrorData(this.m_csElement.GetID(), MuskingumModelDefine.M_KIND, "No Receive Port");
                bRtn = false;
            }
            foreach (var recievePort in csInputCellData)
            {
                if (recievePort.GetReceivePatternID().Equals(MuskingumModelDefine.IN_DISCHARGE_POINT_PTN))
                {
                    // ポイント時系列の場合

                }
                else if (recievePort.GetReceivePatternID().Equals(MuskingumModelDefine.IN_DISCHARGE_1D_PTN))
                {
                    // １次元時系列の場合
                    long lD1 = 0;
                    long lD2 = 0;
                    long lD3 = 0;
                    long lCellDataNum = 0;
                    long lCellDim = recievePort.GetDimension(ref lD1, ref lD2, ref lD3, ref lCellDataNum);

                    if (lCellDim == 1 && lD1 >= 1 && lCellDataNum >= 1)
                    {
                        //do nothing
                    }
                    else
                    {
                        // エラー 
                        bRtn = false;
                        csErrorInf.AddCheckErrorData(this.m_csElement.GetID(), MuskingumModelDefine.M_KIND,
                            "Unexpected receive pattern. (Received from " + m_csInputData[0].GetUpperElementID().ToString() + ")");
                    }
                }

                // エラー (期待している情報ではない)
                else
                {
                    bRtn = false;
                    csErrorInf.AddCheckErrorData(this.m_csElement.GetID(), MuskingumModelDefine.M_KIND,
                        "Unexpected receive pattern. (Received from " + m_csInputData[0].GetUpperElementID().ToString() + ")");
                }

            }

            return bRtn;
        }

        /// <summary><para>method outline:</para>
        /// <para>出力側の接続情報チェック</para>
        /// </summary>
        /// <example><para>usage:</para>
        /// <para>bool bRtn = SendConnectionCheck(ref csErrorInf)</para>
        /// </example>
        /// <param name="csErrorInf">エラー出力</param>
        /// <param name="lOutputDataNum">出力情報数</param>
        /// <param name="csOutputCellData">出力情報配列</param>
        /// <returns>=true:正常、=false:異常</returns>
        /// <exception cref="">無し</exception>
        /// <remarks><para>remarks:</para>
        /// <para>送信端子に設定されている伝送データが自モデルが期待している情報か否かをチェックする</para>
        /// </remarks>
        protected override bool SendConnectionCheck(ref McStructErrorInfo csErrorInf, long lOutputDataNum, McSendCellDataIF[] csOutputCellData)
        {
            bool bRtn = true;

            if (lOutputDataNum == 0)
            {
                // 警告（エラーではない）
                csErrorInf.AddCheckWarningData(this.GetID(), MuskingumModelDefine.M_KIND, "No send port.");
            }
            foreach (var sendPort in csOutputCellData)
            {
                if (sendPort.GetSendPatternID().Equals(MuskingumModelDefine.OUT_DISCHARGE_POINT_PTN))
                {
                    //ポイント時系列の場合

                }
                else if (sendPort.GetSendPatternID().Equals(MuskingumModelDefine.OUT_DISCHARGE_1D_PTN))
                {
                    // 1次元時系列の場合
                    long lD1 = 0;
                    long lD2 = 0;
                    long lD3 = 0;
                    long lCellDataNum = 0;
                    long lCellDim = sendPort.GetDimension(ref lD1, ref lD2, ref lD3, ref lCellDataNum);

                    if (lCellDim == 1 && lD1 >= 1 && lCellDataNum >= 1)
                    {
                        // do nothing
                    }
                    else
                    {
                        bRtn = false;
                        csErrorInf.AddCheckErrorData(this.m_csElement.GetID(), MuskingumModelDefine.M_KIND,
                            "Unexpected receive pattern. (Received from " + m_csInputData[0].GetUpperElementID().ToString() + ")");
                    }
                }
                else if (sendPort.GetSendPatternID().Equals(MuskingumModelDefine.OUT_MODEL_STATE_1D_PTN))
                {
                    // 1次元時系列の場合
                    long lD1 = 0;
                    long lD2 = 0;
                    long lD3 = 0;
                    long lCellDataNum = 0;
                    long lCellDim = sendPort.GetDimension(ref lD1, ref lD2, ref lD3, ref lCellDataNum);

                    if (lCellDim == 1 && lD1 >= 1 && lCellDataNum >= 1)
                    {
                        // do nothing
                    }
                    else
                    {
                        bRtn = false;
                        csErrorInf.AddCheckErrorData(this.m_csElement.GetID(), MuskingumModelDefine.M_KIND,
                            "Unexpected receive pattern. (Received from " + m_csInputData[0].GetUpperElementID().ToString() + ")");
                    }
                }
                else
                {
                    // エラー (期待している情報ではない)
                    bRtn = false;
                    csErrorInf.AddCheckErrorData(this.GetID(), MuskingumModelDefine.M_KIND,
                        "Unexpected send pattern. (Send to " + sendPort.GetLowerElementID().ToString() + ")");
                }
            }

            return bRtn;
        }


        /// <summary><para>method outline:</para>
        /// <para>モデルを初期化する</para>
        /// </summary>
        /// <example><para>usage:</para>
        /// <para>Initialize(csInitialData)</para>
        /// </example>
        /// <param name="csInitialData">初期化設定情報</param>
        /// <param name="lInputDataNum">入力情報数</param>
        /// <param name="csInputCellData">演算に必要な入力情報配列</param>
        /// <returns>無し</returns>
        /// <exception cref="">無し</exception>
        /// <remarks><para>remarks:</para>
        /// <para>無し</para>
        /// </remarks>
        protected override bool Initialize(ref McPropertyInfoRoot csInitialData, long lInputDataNum, ref McReceiveCellDataIF[] csInputCellData)
        {
            var c = m_csCalInfo as MuskingumModelCalInfo;

            //変数初期化
            c.m_DischargeUpper = 0;
            c.m_DischargeLower = 0;
            c.m_Storage = 0;
            c.m_InitStorage = 0;

            var init = csInitialData as McPropertyInfo;
            if (init != null)
            {
                //ダイアログに設定された値を取得
                init.GetInfo("m_InitStorage", ref c.m_InitStorage);
                //変数に反映
                c.m_Storage = c.m_InitStorage;
            }

            //変数正常性のチェック
            McStructErrorInfo csErrorInf = null;
            //HACK: to show error message to user, CheckParameters() is moved from SetProperty()
            //return CheckVariables(ref csErrorInf);
            return CheckParameters(ref csErrorInf) && CheckVariables(ref csErrorInf);
        }

        /// <summary><para>method outline:</para>
        /// <para>モデルを初期化する</para>
        /// </summary>
        /// <example><para>usage:</para>
        /// <para>OnlineInitialize(csInitialData)</para>
        /// </example>
        /// <param name="csInitialData">初期化設定情報</param>
        /// <param name="lInputDataNum">入力情報数</param>
        /// <param name="csInputCellData">演算に必要な入力情報配列</param>
        /// <returns>bool</returns>
        /// <exception cref="">無し</exception>
        /// <remarks><para>remarks:</para>
        /// <para>無し</para>
        /// </remarks>
        protected override bool OnlineInitialize(ref McPropertyInfoRoot csInitialData, long lInputDataNum, ref McReceiveCellDataIF[] csInputCellData)
        {
            // 
            return true;
        }
        //=======================
        // 演算実行処理関連メソッド
        //=======================

        /// <summary><para>method outline:</para>
        /// <para>モデル演算</para>
        /// </summary>
        /// <example><para>usage:</para>
        /// <para>long lRtn = Calculate(ref csInputCellData)</para>
        /// </example>
        /// <param name="lInputDataNum">入力情報数</param>
        /// <param name="csInputCellData">演算に必要な入力情報配列</param>
        /// <returns>=0:正常 -1:異常</returns>
        /// <exception cref="">無し</exception>
        /// <remarks><para>remarks:</para>
        /// <para>無し</para>
        /// </remarks>
        protected override long Calculate(long lInputDataNum, ref McReceiveCellDataIF[] csInputCellData)
        {

            var c = m_csCalInfo as MuskingumModelCalInfo;
            double dischargeUpperNext = 0;

            //流量の受信
            foreach (var recievePort in csInputCellData)
            {
                if (recievePort.GetReceivePatternID().Equals(MuskingumModelDefine.IN_DISCHARGE_POINT_PTN))
                {
                    // ポイント時系列の場合
                    var cellId = recievePort.GetInterpolatedCell(0);  // セル配列の０番目のセルIDを取得
                    dischargeUpperNext += recievePort.Data(cellId, 0);
                }
                else if (recievePort.GetReceivePatternID().Equals(MuskingumModelDefine.IN_DISCHARGE_1D_PTN))
                {
                    // １次元時系列の場合
                    var cellId = recievePort.GetInterpolatedCellD1();  // セル配列の０番目のセルIDを取得
                    dischargeUpperNext += recievePort.Data(cellId[0], 0);
                }
            }

            //下流端流量の算出
            var dischargeLowerNext = c.m_TimeStepInH * c.m_CoefCDenominator * (c.m_DischargeUpper - c.m_DischargeLower)
                + 2.0 * c.m_CoefCDenominator * c.m_Storage
                + (c.m_TimeStepInH - 2.0 * c.m_CoefK * c.m_CoefX) * c.m_CoefCDenominator * dischargeUpperNext;

#if DEBUG
            //検算
            var c0qa2 = c.m_CoefC0 * dischargeUpperNext;
            var c1qa1 = c.m_CoefC1 * c.m_DischargeUpper;
            var c2qb1 = c.m_CoefC2 * c.m_DischargeLower;
            var dischargeLowerNextForCheck = c0qa2 + c1qa1 + c2qb1;
            Debug.Assert(c.m_InitStorage == 0 && Math.Abs(dischargeLowerNext - dischargeLowerNextForCheck) < 0.01 || true);
#endif

            //データの更新
            c.m_DischargeUpper = dischargeUpperNext;
            c.m_DischargeLower = dischargeLowerNext;
            c.m_Storage = c.m_CoefK * (c.m_CoefX * dischargeUpperNext + (1 - c.m_CoefX) * dischargeLowerNext);

            //if (m_csSimTime.GetTime() == m_csStartTime.GetTime())
            //{
            //    Debug.Print("t,m_DischargeUpper,m_DischargeLower");
            //}
            //Debug.Print("{0},{1},{2}", m_csSimTime.GetTime(), c.m_DischargeUpper, c.m_DischargeLower);

            return 0;
        }


        /// <summary><para>method outline:</para>
        /// <para>モデル演算結果を外部のエレメントに対して公開する</para>
        /// </summary>
        /// <example><para>usage:</para>
        /// <para>long lRtn = DataFusion( lOutputDataNum, ref csOutputCellData)</para>
        /// </example>
        /// <param name="lOutputDataNum">出力情報数</param>
        /// <param name="csOutputCellData">出力情報配列</param>
        /// <returns>=0:正常 -1:異常</returns>
        /// <exception cref="">無し</exception>
        /// <remarks><para>remarks:</para>
        /// <para>無し</para>
        /// </remarks>
        protected override long DataFusion(long lOutputDataNum, ref McSendCellDataIF[] csOutputCellData)
        {
            var c = m_csCalInfo as MuskingumModelCalInfo;

            // 出力
            foreach (var sendPort in csOutputCellData)
            {
                if (sendPort.GetSendPatternID().Equals(MuskingumModelDefine.OUT_DISCHARGE_POINT_PTN) ||
                    sendPort.GetSendPatternID().Equals(MuskingumModelDefine.OUT_DISCHARGE_1D_PTN))
                {
                    var cell = sendPort.PrepareSendCellD1();
                    cell[0].m_dData[0] = c.m_DischargeLower;
                }
                else if (sendPort.GetSendPatternID().Equals(MuskingumModelDefine.OUT_MODEL_STATE_1D_PTN))
                {
                    var cell = sendPort.PrepareSendCellD1();
                    int i = 0;
                    if (cell.Length > i)
                    {
                        cell[i].m_dData[0] = c.m_DischargeLower;
                        i++;
                    }
                    if (cell.Length > i)
                    {
                        cell[i].m_dData[0] = c.m_DischargeUpper;
                        i++;
                    }
                    if (cell.Length > i)
                    {
                        cell[i].m_dData[0] = c.m_Storage;
                        i++;
                    }
                }
            }
            return 0;
        }


        //====================
        // その他必要なメソッド
        //====================

        /// <summary><para>method outline:</para>
        /// <para>プロパティ情報を設定する</para>
        /// </summary>
        /// <example><para>usage:</para>
        /// <para>SetProperty(csCellMdlPropertyInfo)</para>
        /// </example>
        /// <param name="csCellMdlPropertyInfo">セル型プロパティ情報</param>
        /// <returns>無し</returns>
        /// <exception cref="">無し</exception>
        /// <remarks><para>remarks:</para>
        /// <para>無し</para>
        /// </remarks>
        public override bool SetProperty(McCellModelPropertyIF csCellMdlPropertyInfo)
        {
            // 使用しやすいようにキャストしておく
            var c = m_csCalInfo as MuskingumModelCalInfo;

            // プロパティ設定
            var prop = csCellMdlPropertyInfo as McCellModelPropertyInfo;
            if (prop != null)
            {
                // 演算ステップ時刻設定
                m_csDltTime = new HySTime(prop.GetStepTime());

                // 各種係数設定
                c.m_TimeStep = prop.GetStepTime();
                c.m_TimeStepInH = c.m_TimeStep / 3600.0;
                prop.GetInfo("m_CoefK", ref c.m_CoefK);
                prop.GetInfo("m_CoefX", ref c.m_CoefX);

                //係数C0, C1, C2の算定
                c.m_CoefCDenominator = 1 / (2.0 * c.m_CoefK * (1.0 - c.m_CoefX) + c.m_TimeStepInH);
                c.m_CoefC0 = (c.m_TimeStepInH - 2.0 * c.m_CoefK * c.m_CoefX) * c.m_CoefCDenominator;
                c.m_CoefC1 = (c.m_TimeStepInH + 2.0 * c.m_CoefK * c.m_CoefX) * c.m_CoefCDenominator;
                c.m_CoefC2 = (2.0 * c.m_CoefK * (1.0 - c.m_CoefX) - c.m_TimeStepInH) * c.m_CoefCDenominator;

                //HACK: to show error message to user, CheckParameters() is moved to Initialize()
                //McStructErrorInfo csErrorInfo = null;
                //return CheckParameters(ref csErrorInf);
                return true;
            }
            else
            {

                return false;
            }
        }

        /// <summary><para>method outline:</para>
        /// <para>演算実行中断中のプロパティ等情報設定</para>
        /// </summary>
        /// <example><para>usage:</para>
        /// <para>bool bRtn = SetOnlineProperty(csCellMdlPropertyInfo)</para>
        /// </example>
        /// <param name="csCellMdlPropertyInfo">プロパティ情報</param>
        /// <returns>無し</returns>
        /// <exception cref="">無し</exception>
        /// <remarks><para>remarks:</para>
        /// <para>演算中断中に　プロパティ情報を変更した場合にコールされる</para>
        /// </remarks>
        public override bool SetOnlineProperty(McCellModelPropertyIF csCellMdlPropertyInfo)
        {
            bool bRtn = false;
            // プロパティ設定
            McCellModelPropertyInfo csPrptyInfo = csCellMdlPropertyInfo as McCellModelPropertyInfo;
            if (csPrptyInfo != null)
            {
                bRtn = true;
            }
            return bRtn;
        }

        private bool CheckVariables(ref McStructErrorInfo csErrorInfo)
        {
            return true;
        }
        private bool CheckParameters(ref McStructErrorInfo csErrorInfo)
        {
            var c = m_csCalInfo as MuskingumModelCalInfo;
            bool ret = true;
            if (Math.Abs(c.m_CoefC0 + c.m_CoefC1 + c.m_CoefC2 - 1) > 0.01)
            {
                //各ceof cの和が1以下であること
                if (csErrorInfo != null)
                    csErrorInfo.AddCheckErrorData(this.m_csElement.GetID(), MuskingumModelDefine.M_KIND, Properties.Resources.STRING_COEFS_OF_C_INVALID);
                McLog.DebugOut(GetSimulationTime(), GetID(), McLog.SYSTEM, " SetProperty Exception Catch", " in (" + this.ToString() + ") :: " + Properties.Resources.STRING_COEFS_OF_C_INVALID); 
                ret = false;
            }
            return ret;
        }
    }

}
