﻿// <summary>ソースコード：モデル要素演算データクラス</summary>
// <author>CommonMP</author>

using System;
using System.Collections.Generic;
using System.Collections;
using System.Text;

using CommonMP.HYSSOP.Interface.HSData;
using CommonMP.HYSSOP.Interface.HSViewer;
using CommonMP.HYSSOP.CoreImpl.HSData;
using CommonMP.HYSSOP.CoreImpl.HSTools;

using CommonMP.HYMCO.Interface.Model;
using CommonMP.HYMCO.Interface.Data;
using CommonMP.HYMCO.Interface.Controller;

namespace jp.or.jccajsce.MuskingumModel
{
    /// <summary><para>class outline:</para>
    /// <para>モデル要素内演算中データクラス</para>
    /// </summary>
    /// <remarks><para>history:</para>
    /// <para>[CommonMP][ver 1.0.0][2010/01][新規作成]</para>
    /// <para>remarks</para>
    /// <para>演算中の状態、初期条件等を保持するデータクラス</para>
    /// </remarks>
    [Serializable]
    public class MuskingumModelCalInfo : McCalInfo
    {
        // parameters
        /// <summary>coef C0 []</summary>
        public double m_CoefC0 = 0;
        /// <summary>coef C1 []</summary>
        public double m_CoefC1 = 0;
        /// <summary>coef C2 []</summary>
        public double m_CoefC2 = 0;
        /// <summary>coef Cの分母 []</summary>
        public double m_CoefCDenominator = 1;
        /// <summary>coef K [hour]</summary>
        public double m_CoefK = 0;
        /// <summary>coef X []</summary>
        public double m_CoefX = 0;

        public double m_TimeStep;
        public double m_TimeStepInH;

        // variables
        public double m_DischargeUpper;
        public double m_DischargeLower;
        public double m_Storage;
        public double m_InitStorage;
        
        /// <summary><para>method outline:</para>
        /// <para>自己複製</para>
        /// </summary>
        /// <example><para>usage:</para>
        /// <para> Clone( ) </para>
        /// </example>
        /// <param name="">無し</param>
        /// <returns>同一内容で　別インスタンス</returns>
        /// <exception cref="">無し</exception>
        /// <remarks><para>remarks:</para>
        /// <para>収束演算時、特定の時刻に状態を戻す場合等に使用する</para>
        /// </remarks>
        public override McCalInfo Clone()
        {
            var csRtn = new MuskingumModelCalInfo();
            csRtn.CopyInfo(this);
            return csRtn;
        }

        /// <summary><para>method outline:</para>
        /// <para>引数で与えられた情報を自分にコピーを行う</para>
        /// </summary>
        /// <example><para>usage:</para>
        /// <para> bool bRtn = CopyInfo(csOrgInfo) </para>
        /// </example>
        /// <param name="csOrgInfo">コピー元情報</param>
        /// <returns> bool true :正常 , false:異常</returns>
        /// <exception cref="">無し</exception>
        /// <remarks><para>remarks:</para>
        /// <para>無し</para>
        /// </remarks>
        public override bool CopyInfo(McCalInfo csOrgInfo)
        {
            var c = (MuskingumModelCalInfo)csOrgInfo;

            //shallow copy
            // parameters
            m_CoefC0 = c.m_CoefC0;
            m_CoefC1 = c.m_CoefC1;
            m_CoefC2 = c.m_CoefC2;
            m_CoefCDenominator = c.m_CoefCDenominator;
            m_CoefK = c.m_CoefK;
            m_CoefX = c.m_CoefX;
            m_TimeStep = c.m_TimeStep;
            m_TimeStepInH = c.m_TimeStepInH;

            m_DischargeUpper = c.m_DischargeUpper;
            m_DischargeLower = c.m_DischargeLower;
            m_Storage = c.m_Storage;
            m_InitStorage = c.m_InitStorage;

            return true;
        }

    }

}
